import {
  AfterViewInit,
  ElementRef,
  HostBinding,
  Input,
  OnInit,
  Output,
  ViewChild,
} from "@angular/core";
import {
  FormBuilder,
  FormGroup,
} from "@angular/forms";
import {SharedAbstractPresentational} from "@app/shared/components/presentationals/shared-abstract/shared-abstract.presentational";
import {Store} from "@ngxs/store";
import {SharedSearchableMultiSelectPresentational} from "@shared/components/presentationals/shared-searchable-multi-select/shared-searchable-multi-select.presentational";
import {SharedSearchableSingleSelectPresentational} from "@shared/components/presentationals/shared-searchable-single-select/shared-searchable-single-select.presentational";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {
  debounceTime,
  distinctUntilChanged,
  tap,
} from "rxjs/operators";
import {
  BaseResourceType,
  isEmptyString,
  isNullOrUndefined,
  MappingObjectUtil,
  MemoizedUtil,
  ObservableUtil,
  QueryParameters,
  ResourceState,
  ResourceStateModel,
} from "solidify-frontend";
import {environment} from "../../../../../environments/environment";

export abstract class SharedSearchableAbstractContentPresentational<TResource extends BaseResourceType> extends SharedAbstractPresentational implements OnInit, AfterViewInit {
  listObs: Observable<TResource[]>;
  totalObs: Observable<number>;
  isLoadingObs: Observable<boolean>;

  form: FormGroup;

  @HostBinding("class.is-loading")
  isLoading: boolean;

  @Input()
  abstract host: SharedSearchableMultiSelectPresentational<ResourceStateModel<TResource>, TResource> | SharedSearchableSingleSelectPresentational<ResourceStateModel<TResource>, TResource>;

  @ViewChild("input", {static: false})
  input: ElementRef;

  private readonly _closeBS: BehaviorSubject<void> = new BehaviorSubject<void>(undefined);
  @Output("closeChange")
  readonly closeObs: Observable<void> = ObservableUtil.asObservable(this._closeBS);

  protected readonly _debounceTime: number = 250;
  protected readonly _pageSize: number = environment.defaultEnumValuePageSizeLazyLoad;

  constructor(protected store: Store,
              protected _fb: FormBuilder) {
    super();
  }

  ngOnInit(): void {
    super.ngOnInit();

    this._initSearchForm();
    this._dispatchInitDataLoad();
    this._retrieveSelector();
    this._observeSearchInput();
    this._observeIsLoading();
  }

  private _initSearchForm(): void {
    this.form = this._fb.group({
      [this.host.formDefinition.search]: [""],
    });
  }

  private _dispatchInitDataLoad(): void {
    this.store.dispatch(new this.host.resourceNameSpace.GetAll(new QueryParameters(this._pageSize)));
  }

  private _retrieveSelector(): void {
    this.listObs = ResourceState.list(this.store, this.host.state);
    this.totalObs = ResourceState.total(this.store, this.host.state);
    this.isLoadingObs = MemoizedUtil.isLoading(this.store, this.host.state);
  }

  private _observeSearchInput(): void {
    this.subscribe(this.form.get(this.host.formDefinition.search).valueChanges.pipe(
      debounceTime(this._debounceTime),
      distinctUntilChanged(),
      tap(v => {
        const queryParameters = new QueryParameters();
        queryParameters.paging.pageSize = this._pageSize;
        MappingObjectUtil.set(queryParameters.search.searchItems, this.host.labelKey, v);
        this.store.dispatch(new this.host.resourceNameSpace.GetAll(queryParameters, true));
      }),
    ));
  }

  ngAfterViewInit(): void {
    this.setFocusOnInput();
  }

  private setFocusOnInput(): void {
    setTimeout(() => {
      this.input.nativeElement.focus();
    }, 0);
  }

  abstract select(value: TResource): void;

  close(): void {
    this._closeBS.next();
  }

  isActive(value: TResource): boolean {
    if (isNullOrUndefined(this.host.formControl.value) || isEmptyString(this.host.formControl.value)) {
      return false;
    }
    return this.host.formControl.value.includes(value[this.host.valueKey]);
  }

  private _observeIsLoading(): void {
    this.subscribe(this.isLoadingObs.pipe(
      distinctUntilChanged(),
      tap(isLoading => this.isLoading = isLoading),
    ));
  }
}



