import {
  FlexibleConnectedPositionStrategy,
  Overlay,
  OverlayRef,
} from "@angular/cdk/overlay";
import {ComponentPortal} from "@angular/cdk/portal";
import {
  AfterViewInit,
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  ComponentRef,
  ElementRef,
  Input,
  OnInit,
  Output,
  ViewContainerRef,
} from "@angular/core";
import {
  ControlValueAccessor,
  FormBuilder,
  FormControl,
  NG_VALUE_ACCESSOR,
} from "@angular/forms";
import {SharedAbstractPresentational} from "@app/shared/components/presentationals/shared-abstract/shared-abstract.presentational";
import {Store} from "@ngxs/store";
import {SharedSearchableMultiSelectContentPresentational} from "@shared/components/presentationals/shared-searchable-multi-select-content/shared-searchable-multi-select-content.presentational";
import {BaseFormDefinition} from "@shared/models/base-form-definition.model";
import _ from "lodash";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {
  distinctUntilChanged,
  filter,
  take,
  tap,
} from "rxjs/operators";
import {
  BaseResourceType,
  FormValidationHelper,
  isEmptyArray,
  isNonEmptyArray,
  isNullOrUndefined,
  MemoizedUtil,
  ObservableUtil,
  PropertyName,
  ResourceNameSpace,
  ResourceState,
  ResourceStateModel,
  StringUtil,
  Type,
} from "solidify-frontend";

@Component({
  selector: "dlcm-shared-searchable-multi-select",
  templateUrl: "./shared-searchable-multi-select.presentational.html",
  styleUrls: ["./shared-searchable-multi-select.presentational.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    {
      provide: NG_VALUE_ACCESSOR,
      multi: true,
      useExisting: SharedSearchableMultiSelectPresentational,
    },
  ],
})
export class SharedSearchableMultiSelectPresentational<TStateModel extends ResourceStateModel<TResource>, TResource extends BaseResourceType> extends SharedAbstractPresentational implements ControlValueAccessor, OnInit, AfterViewInit {
  formDefinition: FormComponentFormDefinition = new FormComponentFormDefinition();

  @Input()
  title: string;

  @Input()
  placeholder: string;

  @Input()
  valueKey: string;

  @Input()
  labelKey: string;

  @Input()
  extraInfoLabelKey: string;

  @Input()
  extraInfoImage: string;

  @Input()
  formControl: FormControl;

  @Input()
  required: boolean;

  @Input()
  resourceNameSpace: ResourceNameSpace;

  @Input()
  state: Type<ResourceState<TStateModel, TResource>>;

  @Input()
  closeAfterValueSelected: boolean = false;

  @Input()
  hideItemSelected: boolean = false;

  listObs: Observable<TResource[]>;
  isLoadingObs: Observable<boolean>;
  selectedItems: TResource[] = [];

  labelInitialized: boolean = false;

  get formValidationHelper(): typeof FormValidationHelper {
    return FormValidationHelper;
  }

  private readonly _valueBS: BehaviorSubject<TResource[] | undefined> = new BehaviorSubject<TResource[] | undefined>(undefined);
  @Output("valueChange")
  readonly valueObs: Observable<TResource[] | undefined> = ObservableUtil.asObservable(this._valueBS);

  constructor(private _store: Store,
              protected _fb: FormBuilder,
              private _elementRef: ElementRef,
              private _viewContainerRef: ViewContainerRef,
              private _overlay: Overlay,
              private _changeDetector: ChangeDetectorRef) {
    super();
  }

  ngOnInit(): void {
    this._manageInitExistingValue();
    // this._managePostExternalChange();
  }

  private _managePostExternalChange(): void {
    this.subscribe(this.formControl.valueChanges.pipe(
      distinctUntilChanged(),
      tap((values: string[]) => {
        values.forEach(v => {
          if (!isNullOrUndefined(this.searchItemInSelectedItemsWithId(v))) {
            return;
          }
          // TODO manage here external add with id (need to retrieve full object in backend)
        });
      }),
    ));
  }

  private _manageInitExistingValue(): void {
    const keys = this.formControl.value;
    if (isNonEmptyArray(keys)) {
      this._dispatchGetActionToRetrieveLabel(keys);
      this._getLabelInitialSelectedValue();
    } else {
      this.labelInitialized = true;
    }
  }

  private _dispatchGetActionToRetrieveLabel(listResId: string[]): void {
    this.listObs = ResourceState.list(this._store, this.state);
    this.isLoadingObs = MemoizedUtil.isLoading(this._store, this.state);
    this._store.dispatch(new this.resourceNameSpace.GetByListId(listResId));
  }

  private _getLabelInitialSelectedValue(): void {
    this.subscribe(this.listObs.pipe(
      distinctUntilChanged(),
      filter((list) => !isNullOrUndefined(list) && !isEmptyArray(list)),
      take(1),
      tap((list) => {
        const keys = this.formControl.value;
        keys.forEach(k => {
          const item = list.find(i => i[this.valueKey] === k);
          this.selectedItems.push(item);
        });
        this.labelInitialized = true;
        this._changeDetector.detectChanges();
      }),
    ));
  }

  propagateChange = (__: any) => {};

  registerOnChange(fn: any): void {
    this.propagateChange = fn;
  }

  registerOnTouched(fn: any): void {
  }

  setDisabledState(isDisabled: boolean): void {
  }

  writeValue(value: string[]): void {
  }

  openOverlay(): void {
    const minWidthField = this._elementRef.nativeElement.getBoundingClientRect().width;
    const overlayRef = this._overlay.create({
      positionStrategy: this._getPosition(this._elementRef),
      scrollStrategy: this._overlay.scrollStrategies.block(),
      hasBackdrop: true,
      backdropClass: "cdk-overlay-transparent-backdrop",
      width: minWidthField,
    });

    const overlayContentComponent = new ComponentPortal(SharedSearchableMultiSelectContentPresentational);
    const componentRef = overlayRef.attach(overlayContentComponent);
    componentRef.instance.host = this;

    this._observeValueSelectedInOverlay(componentRef, overlayRef);
    this._observeCloseEventInOverlay(componentRef, overlayRef);
    this._observeClickBackdrop(componentRef, overlayRef);
  }

  private close(): void {

  }

  private _observeValueSelectedInOverlay(componentRef: ComponentRef<SharedSearchableMultiSelectContentPresentational<TResource>>, overlayRef: OverlayRef): void {
    this.subscribe(componentRef.instance.valueAddedObs.pipe(
      distinctUntilChanged(),
      tap(value => {
        this.add(value);
        this.formControl.markAsTouched();
        if (this.closeAfterValueSelected) {
          this._closeOverlay(overlayRef);
        } else {
          overlayRef.updatePosition();
        }
      }),
    ));

    this.subscribe(componentRef.instance.valueRemovedObs.pipe(
      distinctUntilChanged(),
      tap(value => {
        this.remove(value);
        this.formControl.markAsTouched();
        if (this.closeAfterValueSelected) {
          this._closeOverlay(overlayRef);
        } else {
          overlayRef.updatePosition();
        }
      }),
    ));
  }

  private _observeCloseEventInOverlay(componentRef: ComponentRef<SharedSearchableMultiSelectContentPresentational<TResource>>, overlayRef: OverlayRef): void {
    this.subscribe(componentRef.instance.closeObs.pipe(
      tap(() => {
        this._closeOverlay(overlayRef);
      }),
    ));
  }

  private _observeClickBackdrop(componentRef: ComponentRef<SharedSearchableMultiSelectContentPresentational<TResource>>, overlayRef: OverlayRef): void {
    this.subscribe(overlayRef.backdropClick().pipe(
      tap(() => {
        this._closeOverlay(overlayRef);
      }),
    ));
  }

  private _closeOverlay(overlayRef: OverlayRef): void {
    overlayRef.detach();
    this.formControl.markAsTouched();
    this._changeDetector.detectChanges();
  }

  private _getPosition(elementToConnectTo: ElementRef): FlexibleConnectedPositionStrategy {
    return this._overlay.position()
      .flexibleConnectedTo(elementToConnectTo)
      .withFlexibleDimensions(false)
      .withPositions([
        {
          originX: "start",
          originY: "bottom",
          overlayX: "start",
          overlayY: "top",
        },
      ]);
  }

  focusInputIfNotClickTarget($event: MouseEvent, input: HTMLInputElement): void {
    if ($event.target !== input) {
      input.focus();
    }
  }

  removable: boolean = true;

  add(item: TResource): void {
    if (!isNullOrUndefined(this.searchItemInSelectedItems(item))) {
      return;
    }
    this.selectedItems.push(item);
    this._changeDetector.detectChanges();
    this._updateFormControlWithSelectedItem();
  }

  remove(item: TResource): void {
    const index = this.searchIndexItemInSelectedItems(item);
    if (index === -1) {
      return;
    }
    this.selectedItems.splice(index, 1);
    this._updateFormControlWithSelectedItem();
    this._changeDetector.detectChanges();
  }

  private _updateFormControlWithSelectedItem(): void {
    const arrayKeys = _.map(this.selectedItems, this.valueKey);
    this.formControl.setValue(arrayKeys);
    this.propagateChange(arrayKeys);
    this._valueBS.next(arrayKeys);
  }

  private searchItemInSelectedItemsWithId(id: string): TResource | undefined {
    const index = this.searchIndexItemInSelectedItems({[this.valueKey]: id} as any);
    if (index === -1) {
      return this.selectedItems[index];
    }
  }

  private searchItemInSelectedItems(item: TResource): TResource | undefined {
    const index = this.searchIndexItemInSelectedItems(item);
    if (index === -1) {
      return undefined;
    }
    return this.selectedItems[index];
  }

  private searchIndexItemInSelectedItems(item: TResource): number {
    return this.selectedItems.findIndex(s => s[this.valueKey] === item[this.valueKey]);
  }

  getTooltip(value: TResource): string {
    let extraInfo = StringUtil.stringEmpty;
    if (!isNullOrUndefined(this.extraInfoLabelKey) && !isNullOrUndefined(value[this.extraInfoLabelKey])) {
      extraInfo = " - " + value[this.extraInfoLabelKey];
    }
    return value[this.labelKey] + extraInfo;
  }
}

class FormComponentFormDefinition extends BaseFormDefinition {
  @PropertyName() search: string;
  @PropertyName() value: string;
}
